/*
 * Generic hash table handler...
 *
 * Copyright 1998 by Gray Watson.
 *
 * This file is part of the install package.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and that the name of Gray Watson not be used in advertising or
 * publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted at gray@letters.com
 *
 * $Id: install.c,v 1.42 2000/03/07 20:46:10 gray Exp $
 */

#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#if HAVE_STDLIB_H
# include <stdlib.h>
#endif
#if HAVE_STRING_H
# include <string.h>
#endif
#if HAVE_UNISTD_H
#  include <unistd.h>
#endif
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include "argv.h"
#include "conf.h"

static	char	*rcs_id =
  "$Id: install.c,v 1.42 2000/03/07 20:46:10 gray Exp $";

/*
 * NOTE: the version numbers here should match those in install.c,
 * ChangeLog, and NEWS.
 */
static	char	*version_id = "$Install Version: 4.0.0 March-7-2000 $";

/* unlink_file() return codes */
#define UNLINK_ERROR_NONE	1	/* unlink succeeded */
#define UNLINK_ERROR_ENOENT	2	/* file did not exist */
#define UNLINK_ERROR_NOCONFIRM	3	/* unlink was not confirmed */
#define UNLINK_ERROR_GENTLE	4	/* gentle on and file exists */
#define UNLINK_ERROR_OTHER	5	/* other error */

/*
 * install defines
 */
#define DEFAULT_MODE		0555	/* default file installation mode */
#define DEFAULT_MODE_STR	"0555"	/* for usage messages */
#define DEFAULT_DIRMODE		0755	/* default dir installation mode */
#define DEFAULT_DIRMODE_STR	"0755"	/* for usage messages */
#define DEFAULT_GROUP		"bin"	/* default installation group */
#define DEFAULT_OWNER		"bin"	/* default installation owner */

#define ARG_GROUP		'g'	/* group argument */
#define ARG_MODE		'm'	/* file mode argument */
#define ARG_DIR_MODE		'M'	/* directory mode argument */
#define ARG_OWNER		'o'	/* owner argument */

#define ROOT_UID		0	/* super-user id */
#define BACKUP_EXTENSION	".bak"	/* for -backup flag */

#define BUFFER_SIZE		(10 * 1024)	/* size of the copy buffer */

/* local variables */
static	char		buffer[BUFFER_SIZE];	/* copy buffer */
static	struct group	*group_p = NULL;	/* group info structure */
static	struct passwd	*passwd_p = NULL;	/* owner info structure */

/* argument variables */
static	int		copy_files_b = ARGV_FALSE;	/* cp file not mv */
static	int		link_files_b = ARGV_FALSE;	/* ln file not mv */
#ifdef HAVE_SYMLINK
static	int		rel_symlink_files_b = ARGV_FALSE; /* ln -s file */
static	int		abs_symlink_files_b = ARGV_FALSE; /* ln -s file abs */
#endif
static	int		dir_make_b = ARGV_FALSE;	/* make a directory */

static	int		backup_files_b = ARGV_FALSE;	/* don't unlink, bak */
static	int		compare_b = ARGV_FALSE;		/* compare after */
static	int		dir_makeall_b = ARGV_FALSE;	/* don't unlink, bak */
static	int		examine_b = ARGV_FALSE;		/* examine b4 install*/
static	int		error_code_b = ARGV_FALSE;	/* no install error */
static	char		*file_group = DEFAULT_GROUP;	/* group of file */
static	int		file_mode = DEFAULT_MODE;	/* mode of files */
static	int		dir_mode = DEFAULT_DIRMODE;	/* mode of dirs */
static	int		execute_cmds_b = ARGV_TRUE;	/* execute commands */
static	char		*file_owner = DEFAULT_OWNER;	/* owner of file */
static	int		gentle_b = ARGV_FALSE;		/* be gentle, no rm */
static	int		help_b = ARGV_FALSE;		/* display help mess */
static	int		interactive_b = ARGV_FALSE;	/* ask before unlink */
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(RANLIB_PROG)
static	int		ranlib_libs_b = ARGV_FALSE;	/* ranlib file after */
#endif
static	int		remove_ext_b = ARGV_FALSE;	/* remove extension */
static	int		quiet_b	 = ARGV_FALSE;		/* quiet mode */
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(STRIP_PROG)
static	int		strip_bins_b = ARGV_FALSE;	/* strip file after */
#endif
static	int		sparse_b = ARGV_FALSE;		/* sparsely install */
#ifdef HAVE_UMASK
static	int		umask_mode = -1;		/* default umask */
#endif
static	int		verbose_b = ARGV_FALSE;		/* verbose flag */
static	int		very_verbose_b = ARGV_FALSE;	/* very-verbose flag */
static	argv_array_t	files;				/* work files/dirs */

/* argument array */
static	argv_t	args[] = {
  { 'c',	"copy-files",	ARGV_BOOL_INT,	&copy_files_b,
    NULL,				"copy file(s), don't move" },
  { ARGV_OR },
  { 'l',	"link-files",	ARGV_BOOL_INT,	&link_files_b,
    NULL,				"hard link file(s), don't move" },
#ifdef HAVE_SYMLINK
  { ARGV_OR },
  { 'p',	"symlink-relative", ARGV_BOOL_INT,	&rel_symlink_files_b,
    NULL,				"relative symlink file(s)" },
  { ARGV_OR },
  { 'P',	"symlink-absolute", ARGV_BOOL_INT,	&abs_symlink_files_b,
    NULL,				"absolute symlink file(s)" },
#endif
  { ARGV_OR },
  { 'd',	"make-dir",	ARGV_BOOL_INT,	&dir_make_b,
    NULL,		"make a directory" },
  
  { 'b',	"backup",	ARGV_BOOL_INT,	&backup_files_b,
    NULL,		"backup destination" },
  { 'C',	"compare",	ARGV_BOOL_INT,	&compare_b,
    NULL,		"compare files after install" },
  { 'D',	"mkdir-all",	ARGV_BOOL_INT,	&dir_makeall_b,
    NULL,		"make all directories to target" },
  { 'e',	"examine",	ARGV_BOOL_INT,	&examine_b,
    NULL,		"examine files before install" },
  { 'E',	"error-code",	ARGV_BOOL_INT,	&error_code_b,
    NULL,		"generate error if no install" },
  { ARG_GROUP,	"group-id",	ARGV_CHAR_P,	&file_group,
    "group",		"group id name (default " DEFAULT_GROUP ")" },
  { 'G',	"gentle",	ARGV_BOOL_INT,	&gentle_b,
    NULL,		"be gentle, don't unlink dests" },
  { 'h',	"help",		ARGV_BOOL_INT,	&help_b,
    NULL,		"display help message" },
  { 'i',	"interactive",	ARGV_BOOL_INT,	&interactive_b,
    NULL,		"ask before removing files" },
  { ARG_MODE,	"file-mode",	ARGV_OCT,	&file_mode,
    "octal-mode",	"create file mode (default " DEFAULT_MODE_STR ")" },
  { ARG_DIR_MODE, "dir-mode",	ARGV_OCT,	&dir_mode,
    "octal-mode",	"create dir mode (default " DEFAULT_DIRMODE_STR ")" },
  { 'n',	 "no-execute",	ARGV_BOOL_INT_NEG, &execute_cmds_b,
    NULL,		"do not execute any commands" },
  { ARG_OWNER,	"owner-id",	ARGV_CHAR_P,	&file_owner,
    "owner",		"owner id name (default " DEFAULT_OWNER ")" },
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(RANLIB_PROG)
  { 'r',	"ranlib",	ARGV_BOOL_INT,	&ranlib_libs_b,
    NULL,		"ranlib destination library" },
#endif
  { 'R',	"remove-ext",	ARGV_BOOL_INT,	&remove_ext_b,
    NULL,		"remove extension to install" },
  { 'q',	"quiet",	ARGV_BOOL_INT,	&quiet_b,
    NULL,		"quiet and don't complain as much" },
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(STRIP_PROG)
  { 's',	"strip",	ARGV_BOOL_INT,	&strip_bins_b,
    NULL,		"strip destination binary" },
#endif
  { 'S',	"sparse",	ARGV_BOOL_INT,	&sparse_b,
    NULL,		"install files sparse-ly" },
#ifdef HAVE_UMASK
  { 'u',	"umask-mode",	ARGV_OCT,	&umask_mode,
    "octal-mode",	"mode to set umask to" },
#endif
  { 'v',	"verbose",	ARGV_BOOL_INT,	&verbose_b,
    NULL,		"verbose mode" },
  { 'V',	"very-verbose",	ARGV_BOOL_INT,	&very_verbose_b,
    NULL,		"very-verbose mode" },
  
  { ARGV_MAYBE,	NULL,		ARGV_CHAR_P | ARGV_FLAG_ARRAY,	&files,
    "[file(s)] dir",	"files to install or mkdir arg" },
  { ARGV_LAST }
};

/*
 * static void make_path
 *
 * DESCRIPTION:
 *
 * Make all the directories to a certain path.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * path -> Path to which we need to make all of the directories.
 */
static	void	make_path(const char *path)
{
  char	buf[MAXPATHLEN], *buf_p;
  
  if (very_verbose_b) {
    (void)printf("  making all directories to '%s'\n", path);
  }
  
  (void)strcpy(buf, path);
  
  /* ignore 1st bunch of / because / always exists */
  buf_p = buf;
  for (buf_p = buf; *buf_p == '/'; buf_p++) {
  }
  
  for (; *buf_p != '\0'; buf_p++) {
    if (*buf_p != '/') {
      continue;
    }
    
    /* punch the null here - will return the / below */
    *buf_p = '\0';
    if ((! execute_cmds_b) || mkdir(buf, dir_mode) == 0) {
      if (very_verbose_b) {
	(void)printf("  created directory '%s', mode %#o\n", buf, dir_mode);
      }
      else if (verbose_b) {
	(void)printf("Created directory '%s'\n", buf);
      }
    }
    /*
     * NOTE: we check for access error because NFS reports access if
     * you try to make an existing directory.  Doh!
     */
    else if (errno != EEXIST && errno != EACCES) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: could not create directory '%s': %s\n",
		      argv_program, buf, strerror(errno));
      }
      exit(1);
    }
    *buf_p = '/';
    /* skip any multiple /'s */
    for (; *(buf_p + 1) == '/'; buf_p++) {
    }
  }
}

/*
 * static void sparse_write
 *
 * DESCRIPTION:
 *
 * Write a buffer of bytes to a file while lseek-ing over sections of
 * zero bytes to preserve the files sparse-ness.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * fd -> File descriptor we are writing.
 *
 * path -> Path associated with the fd.
 *
 * buf -> Buffer of bytes to write to the file descriptor.  If NULL then
 * we are writing the last block of '\0'.
 *
 * byte_n -> Number of bytes in the buffer.
 */
static	void	sparse_write(const int fd, const char *path,
			     const char *buf, const int byte_n)
{
  static int	null_c = 0, good_c = 0;
  const char	*buf_p;
  char		null_buf[1] = { '\0' };
  
  /* cleanup the nulls? */
  if (buf == NULL) {
    if (null_c > 0) {
      /* we need to physically write() a null at the end */
      null_c--;
      
      if (null_c > 0) {
	if (execute_cmds_b && lseek(fd, null_c, SEEK_CUR) <= 0) {
	  if (! quiet_b) {
	    (void)fprintf(stderr,
			  "%s: lseeking forward %d bytes in '%s': %s\n",
			  argv_program, null_c, path, strerror(errno));
	  }
	  exit(1);
	}
      }
      
      /* write the last null */
      if (execute_cmds_b && write(fd, null_buf, 1) != 1) {
	if (! quiet_b) {
	  (void)fprintf(stderr, "%s: writing to '%s': %s\n",
			argv_program, path, strerror(errno));
	}
	exit(1);
      }
      
      null_c = 0;
    }
    
    return;
  }
  
  /* run through the buffer watching the nulls */
  for (buf_p = buf; buf_p < buf + byte_n; buf_p++) {
    if (*buf_p == '\0') {
      if (good_c > 0) {
	if (execute_cmds_b && write(fd, buf_p - good_c, good_c) != good_c) {
	  if (! quiet_b) {
	    (void)fprintf(stderr, "%s: writing to '%s': %s\n",
			  argv_program, path, strerror(errno));
	  }
	  exit(1);
	}
	good_c = 0;
      }
      
      null_c++;
    } else {
      if (null_c > 0) {
	if (execute_cmds_b && lseek(fd, null_c, SEEK_CUR) < 0) {
	  if (! quiet_b) {
	    (void)fprintf(stderr,
			  "%s: lseeking forward %d bytes in '%s': %s\n",
			  argv_program, null_c, path, strerror(errno));
	  }
	  exit(1);
	}
	null_c = 0;
      }
      
      good_c++;
    }
  }
  
  /*
   * any good chars left then write since we need to re-use the buffer.
   * the nulls get cleaned up when a NULL buffer is sent in above
   */
  if (good_c > 0) {
    if (execute_cmds_b && write(fd, buf_p - good_c, good_c) != good_c) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: writing to '%s': %s\n",
		      argv_program, path, strerror(errno));
      }
      exit(1);
    }
    
    good_c = 0;
  }
}

/*
 * static void copy_file
 *
 * DESCRIPTION:
 *
 * Copy a file from source to destination.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * from -> Source path.
 *
 * to -> Destination path.
 *
 * mode -> Mode to create the file with.
 */
static	void	copy_file(const char *from, const char *to, const int mode)
{
  int	from_fd, to_fd, ret;
  
  if (very_verbose_b) {
    (void)printf("  %scopying '%s' -> '%s'\n",
		 (sparse_b ? "sparse " : ""), from, to);
  }
  
  if (! execute_cmds_b) {
    return;
  }
  
  /* open 'from' file */
  from_fd = open(from, O_RDONLY, 0);
  if (from_fd < 0) {
    if (! quiet_b) {
      (void)fprintf(stderr, "%s: opening '%s': %s\n",
		    argv_program, from, strerror(errno));
    }
    exit(1);
  }
  
  /* open 'to' file */
  to_fd = open(to, O_WRONLY | O_CREAT, mode);
  if (to_fd < 0) {
    if (errno == ENOENT && dir_makeall_b) {
      make_path(to);
      to_fd = open(to, O_WRONLY | O_CREAT, mode);
    }
    if (to_fd < 0) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: opening '%s': %s\n",
		      argv_program, to, strerror(errno));
      }
      exit(1);
    }
  }
  
  /* read in buffers and write to 'to' */
  for (;;) {
    ret = read(from_fd, buffer, BUFFER_SIZE);
    if (ret < 0) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: reading from '%s': %s\n",
		      argv_program, from, strerror(errno));
      }
      exit(1);
    }
    if (ret == 0) {
      break;
    }
    
    if (sparse_b) {
      sparse_write(to_fd, to, buffer, ret);
    }
    else if (write(to_fd, buffer, ret) != ret) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: writing to '%s': %s\n",
		      argv_program, to, strerror(errno));
      }
      exit(1);
    }
  }
  
  /* write the last block of stuff to the file */
  if (sparse_b) {
    sparse_write(to_fd, to, NULL, 0);
  }
  
  /* close files */
  (void)close(from_fd);
  (void)close(to_fd);
}

/*
 * static int unlink_file
 *
 * DESCRIPTION:
 *
 * Unlink (i.e. remove) a path.
 *
 * RETURNS:
 *
 * Success - UNLINK_ERROR_NONE
 *
 * Failure - Unlink-file return code
 *
 * ARGUMENTS:
 *
 * path -> Path to unlink.
 *
 * force_b -> Whether we obey the gentle flag or force the unlink.
 */
static	int	unlink_file(const char *path, const int force_b)
{
  if (very_verbose_b) {
    (void)printf("  unlinking '%s'\n", path);
  }
  
  if ((! force_b) && gentle_b) {
    if (! quiet_b) {
      (void)fprintf(stderr, "%s: file '%s' exists\n", argv_program, path);
    }
    return UNLINK_ERROR_GENTLE;
  }
  
  /* should we actually do this? */
  if (interactive_b) {
    char	str[80];
    
    (void)printf("Should %s remove '%s'? ", argv_program, path);
    
    if (fgets(str, sizeof(str), stdin) == NULL
	|| (str[0] != 'y' && str[0] != 'Y')) {
      return UNLINK_ERROR_NOCONFIRM;
    }
  }
  
  if (execute_cmds_b && unlink(path) != 0) {
    if (errno == ENOENT) {
      return UNLINK_ERROR_ENOENT;
    }
    else {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: unlinking '%s': %s\n",
		      argv_program, path, strerror(errno));
      }
      return UNLINK_ERROR_OTHER;
    }
  }
  
  return UNLINK_ERROR_NONE;
}

#ifdef HAVE_SYMLINK
/*
 * static void reduce_path
 *
 * DESCRIPTION:
 *
 * Remove ./ prefix, any multiple slashes (//), and take care of relative
 * references inside of a path.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * path <-> Path that we are reducing.
 */
static	void	reduce_path(char *path)
{
  char	*path_p, *new_p;
  
  path_p = path;
  new_p = path;
  
  /*
   * Absorb any heading "./".  We can't do this inside the loop
   * because the '/' is the trigger character not '.'.
   */
  while (*path_p == '.' && *(path_p + 1) == '/') {
    path_p += 2;
    /* absorb more / */
    for (; *path_p == '/'; path_p++) {
    }
  }
  
  while (*path_p != '\0') {
    
    /* normal character */
    if (*path_p != '/') {
      *new_p++ = *path_p++;
      continue;
    }
    
    *new_p++ = *path_p++;
    
    /* absorb multiple / */
    for (; *path_p == '/'; path_p++) {
    }
    
    /* absorb any "./" */
    while (*path_p == '.' && *(path_p + 1) == '/') {
      path_p += 2;
      /* again, absorb more / */
      for (; *path_p == '/'; path_p++) {
      }
    }
    
    /* do we have a "/../" */
    while (*path_p == '.' && *(path_p + 1) == '.' && *(path_p + 2) == '/') {
      path_p += 3;
      /* again, absorb more / */
      for (; *path_p == '/'; path_p++) {
      }
      
      /* move the new path up a directory level */
      for (new_p -= 2; new_p >= path && *new_p != '/'; new_p--) {
      }
      if (new_p < path) {
	(void)fprintf(stderr,
		      "%s: tried to reference .. from the root in '%s'\n",
		      argv_program, path);
	exit(1);
      }
      new_p++;
    }
  }
  *new_p = '\0';
}

/*
 * static void make_relative
 *
 * DESCRIPTION:
 *
 * Build the relative path from the destination path back to the source
 * path.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * from -> Destination absolute path.
 *
 * to -> Source absolute path.
 *
 * dest <- Destination buffer to save the new relative path.
 */
static	void	make_relative(const char *from, const char *to, char *dest)
{
  const char	*from_p, *to_p;
  char		*dest_p;
  int		dir_c;
  
  from_p = from;
  to_p = to;
  
  /* count the common directories */
  dir_c = 0;
  while (*from_p != '\0' && *to_p != '\0') {
    if (*from_p != *to_p) {
      break;
    }
    if (*from_p == '/') {
      dir_c++;
    }
    from_p++;
    to_p++;
  }
  
  dest_p = dest;
  
  /* each directory left in from, means we need to go up a directory */
  for (; *from_p != '\0'; from_p++) {
    if (*from_p == '/') {
      *dest_p++ = '.';
      *dest_p++ = '.';
      *dest_p++ = '/';
    }
  }
  
  /* move to-p back to the last / */
  for (; *to_p != '/'; to_p--) {
  }
  
  /* skip the 1st / and copy the rest of the to into the dest */
  for (to_p++; *to_p != '\0'; to_p++) {
    *dest_p++ = *to_p;
  }
  *dest_p = '\0';
}

/*
 * static void symlink_file
 *
 * DESCRIPTION:
 *
 * Symlink a file from the source path to the destination path.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * from - Source path.
 *
 * to - Destination path.
 */
static	void	symlink_file(const char *from, const char *to)
{
  char	pwd[MAXPATHLEN], from_path[MAXPATHLEN], *from_p;
  char	to_path[MAXPATHLEN], *to_p;
  char	dest[MAXPATHLEN];
  int	len;
  
  if (from[0] == '/') {
    from_p = (char *)from;
  }
  else {
    /* make sure we have enough for "pwd/from" */
    len = MAXPATHLEN - 1 - strlen(from);
    
    /* get our current path */
    getcwd(pwd, len);
    
    /*
     * Create our absolute from path.  If the from was specified as
     * ../blah then yes we will have a .. in our path.
     */
    (void)sprintf(from_path, "%s/%s", pwd, from);
    reduce_path(from_path);
    from_p = from_path;
  }
  
  /* if we are dealing with relative paths then determine the path back */
  if (rel_symlink_files_b) {
    if (to[0] == '/') {
      to_p = (char *)to;
    }
    else {
      (void)sprintf(to_path, "%s/%s", pwd, to);
      reduce_path(to_path);
      to_p = to_path;
    }
    
    /* make the relative path */
    make_relative(to_p, from_p, dest);
    from_p = dest;
  }
  
  if (very_verbose_b) {
    (void)printf("  sym-linking '%s' to point to '%s'\n", to, from_p);
  }
  
  /* actually do the symlink */
  if ((! execute_cmds_b) || symlink(from_p, to) == 0) {
    return;
  }
  
  if (errno == ENOENT && dir_makeall_b) {
    make_path(to);
    if (execute_cmds_b && symlink(from_p, to) == 0) {
      return;
    }
  }
  
  if (very_verbose_b) {
    (void)printf("    failed: %s\n", strerror(errno));
  }
  else if (! quiet_b) {
    (void)fprintf(stderr, "%s: symlinking '%s' to '%s': %s\n",
		  argv_program, from, to, strerror(errno));
  }
  exit(1);
}
#endif

/*
 * static int link_file
 *
 * DESCRIPTION:
 *
 * Link the source path to the destination path path.  
 *
 * RETURNS:
 *
 * Success - 1
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * from - Source path.
 *
 * to - Destination path.
 */
static	int	link_file(const char *from, const char *to)
{
  if (very_verbose_b) {
    (void)printf("  linking '%s' to '%s'\n", from, to);
  }
  
  /* link the to file to the backup file */
  if ((! execute_cmds_b) || link(from, to) == 0) {
    return 1;
  }
  
  if (errno == ENOENT && dir_makeall_b) {
    make_path(to);
    if ((! execute_cmds_b) || link(from, to) == 0) {
      return 1;
    }
  }
  
  if (very_verbose_b) {
    (void)printf("    failed: %s\n", strerror(errno));
  }
  
  return 0;
}

/*
 * static int install_file
 *
 * DESCRIPTION:
 *
 * Move/copy file FROM -> TO.  If TO_EXIST then do some extra stuff.
 *
 * RETURNS:
 *
 * Success - 1 if the file was installed
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * from -> Source path.
 *
 * to -> Destination path.
 *
 * to_exist_b -> Set to 1 if the destination file already exists else 0.
 */
static	int	install_file(const char *from, const char *to,
			     const int to_exist_b)
{
  int	file_linked_b = 0;
  char	to_bak[MAXPATHLEN];
  int	unlink_err;
  
  if (verbose_b || very_verbose_b) {
    (void)printf("Installing '%s' as '%s'\n", from, to);
  }
  
  /* backup the 'to' file */
  if (backup_files_b && to_exist_b) {
    
    /* create the backup path */
    (void)sprintf(to_bak, "%s%s", to, BACKUP_EXTENSION);
    
    /* should we force this unlink? */
    unlink_err = unlink_file(to_bak, 0);
    if (unlink_err != UNLINK_ERROR_NONE && unlink_err != UNLINK_ERROR_ENOENT) {
      return 0;
    }
    
    /* link the to file to the backup file */
    if (link_file(to, to_bak) != 1) {
      if ((! very_verbose_b) && (! quiet_b)) {
	(void)fprintf(stderr, "%s: linking from '%s' to '%s' failed: %s\n",
		      argv_program, to, to_bak, strerror(errno));
      }
      return 0;
    }
  }
  
  /* remove the to path */
  unlink_err = 0;
  if (to_exist_b) {
    unlink_err = unlink_file(to, 0);
    if (unlink_err != UNLINK_ERROR_NONE && unlink_err != UNLINK_ERROR_ENOENT) {
      /* forcably unlink any created backup file */
      if (backup_files_b) {
	(void)unlink_file(to_bak, 1);
      }
      return 0;
    }
  }
  
#ifdef HAVE_SYMLINK
  /* are we sym-linking the files */
  if (rel_symlink_files_b || abs_symlink_files_b) {
    symlink_file(from, to);
    return 1;
  }
#endif
  
  /* try and link first */
  if (! copy_files_b) {
    if (link_file(from, to)) {
      file_linked_b = 1;
    }
    else if (errno != EXDEV) {
      if ((! very_verbose_b) && (! quiet_b)) {
	(void)fprintf(stderr, "%s: linking from '%s' to '%s': %s\n",
		      argv_program, from, to, strerror(errno));
      }
      if (unlink_err == UNLINK_ERROR_NONE && (! quiet_b)) {
	(void)fprintf(stderr,
		      "%s: WARNING: destination unlinked but not re-linked\n",
		      argv_program);
      }
      return 0;
    }
  }
  
  if (! file_linked_b) {
    copy_file(from, to, 0600);
  }
  
  /*
   * forcably unlink the 'from' file if not copying or linking
   */
  if ((! copy_files_b) && (! link_files_b)
#ifdef HAVE_SYMLINK
      && (! rel_symlink_files_b)
      && (! abs_symlink_files_b)
#endif
      ) {
    unlink_err = unlink_file(from, 1);
    if (unlink_err != UNLINK_ERROR_NONE && unlink_err != UNLINK_ERROR_ENOENT) {
      return 0;
    }
  }
  
  return 1;
}

/*
 * static int set_permissions
 *
 * DESCRIPTION:
 *
 * Set file permissions for a path.
 *
 * RETURNS:
 *
 * Success - 1 if anything changed
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * path -> File path whose permissions we are changing.
 *
 * path_stat_p -> Permissions we are setting.
 */
static	int	set_permissions(const char *path,
				const struct stat *path_stat_p)
{
  int	final = 0;
  
  /*
   * set file owner and group
   */
  
  if (path_stat_p == NULL
      || path_stat_p->st_uid != passwd_p->pw_uid
      || path_stat_p->st_gid != passwd_p->pw_gid) {
#ifdef HAVE_CHOWN
    if ((! execute_cmds_b)
	|| chown(path, passwd_p->pw_uid, group_p->gr_gid) == 0) {
      final = 1;
      if (very_verbose_b) {
	(void)printf("  owner/group set to '%s.%s'\n",
		     passwd_p->pw_name, group_p->gr_name);
      }
    }
    else {
      if (errno != EPERM) {
	if (! quiet_b) {
	  (void)fprintf(stderr,
			"%s: changing owner/group of '%s' to '%s.%s': %s\n",
			argv_program, path, passwd_p->pw_name,
			group_p->gr_name, strerror(errno));
	}
      }
      else if (very_verbose_b) {
	(void)printf("  could not change owner/group to '%s.%s'\n",
		     passwd_p->pw_name, group_p->gr_name);
      }
    }
#endif /* HAVE_CHOWN */
  }
  else
    if (very_verbose_b) {
      (void)printf("  owner/group already '%s.%s'\n",
		   passwd_p->pw_name, group_p->gr_name);
    }
  
  /*
   * set file mode
   */
  if (path_stat_p == NULL || (path_stat_p->st_mode & 07777) != file_mode) {
#ifdef HAVE_CHMOD
    if ((! execute_cmds_b)
	|| chmod(path, file_mode) == 0) {
      final = 1;
      if (very_verbose_b) {
	(void)printf("  mode set to '%#o'\n", file_mode);
      }
    }
    else if (! quiet_b) {
      (void)fprintf(stderr, "%s: could not change mode on '%s' to '%#o': %s\n",
		    argv_program, path, file_mode, strerror(errno));
    }
#endif /* HAVE_CHMOD */
  }
  else {
    if (very_verbose_b) {
      (void)printf("  mode already '%#o'\n", file_mode);
    }
  }
  
  return final;
}

/*
 * static int run_command
 *
 * DESCRIPTION:
 *
 * Run a command with a single argument
 *
 * RETURNS:
 *
 * Success - 1
 *
 * Failure - 0
 *
 * ARGUMENTS:
 *
 * prog -> Name of the program we are to run.
 *
 * arg -> 1st argument to pass into the program.
 */
static	int	run_command(const char *prog, const char *arg)
{
  int	pid, code;
  
  if (very_verbose_b) {
    (void)printf("  running command '%s %s'\n", prog, arg);
  }
  
  if (! execute_cmds_b) {
    return 1;
  }
  
#if HAVE_FORK
  pid = fork();
#else
  pid = -1;
#endif
  if (pid < 0) {
    (void)fprintf(stderr, "%s: could not fork to run program '%s': %s\n",
		  argv_program, prog, strerror(errno));
    exit(1);
  }
  
  /* if it is the child then run the program */
  if (pid == 0) {
#if HAVE_EXECLP
    (void)execlp(prog, prog, arg, NULL);
#endif
    (void)fprintf(stderr, "%s: could not execlp program '%s': %s\n",
		  argv_program, prog, strerror(errno));
    exit(1);
  }
  
  /* did wait find the right kid and were there no errors? */
  if (wait(&code) != pid || code != 0) {
    if (very_verbose_b) {
      (void)fprintf(stderr, "%s: program '%s' failed with a code %d\n",
		    argv_program, prog, code);
    }
    return 0;
  }
  
  return 1;
}

/*
 * static int are_name
 *
 * DESCRIPTION:
 *
 * Compare two files to see if they contain the same bytes.
 *
 * RETURNS:
 *
 * 1 if the files are the same, else 0.
 *
 * ARGUMENTS:
 *
 * from -> First file.
 *
 * to -> Second file.
 */
static	int	are_same(const char *from, const char *to)
{
  char	buffer2[BUFFER_SIZE];
  int	fd1, fd2, ret, len;
  int	same_b = 1;
  
  if (very_verbose_b) {
    (void)printf("Examining '%s' and '%s'\n", from, to);
  }
  
  /* we assume that file 1 exists */
  fd1 = open(from, O_RDONLY, 0);
  if (fd1 < 0) {
    if (! quiet_b) {
      (void)fprintf(stderr, "%s: opening '%s': %s\n",
		    argv_program, from, strerror(errno));
    }
    exit(1);
  }
  /* we assume that file 2 exists */
  fd2 = open(to, O_RDONLY, 0);
  if (fd2 < 0) {
    if (! quiet_b) {
      (void)fprintf(stderr, "%s: opening '%s': %s\n",
		    argv_program, to, strerror(errno));
    }
    exit(1);
  }
  
  for (;;) {
    len = read(fd1, buffer, BUFFER_SIZE);
    if (len < 0) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: reading from '%s': %s\n",
		      argv_program, from, strerror(errno));
      }
      exit(1);
    }
    
    ret = read(fd2, buffer2, BUFFER_SIZE);
    if (ret < 0) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: reading from '%s': %s\n",
		      argv_program, to, strerror(errno));
      }
      exit(1);
    }
    
    if (ret != len) {
      same_b = 0;
      break;
    }
    
    /* EOF reached? */
    if (len == 0) {
      break;
    }
    
    /* compare each character */
    if (memcmp(buffer, buffer2, len) != 0) {
      same_b = 0;
      break;
    }
  }
  
  (void)close(fd1);
  (void)close(fd2);
  
  if (very_verbose_b) {
    if (same_b) {
      (void)printf("  file '%s' and '%s' are the same\n", from, to);
    }
    else {
      (void)printf("  file '%s' and '%s' are different\n", from, to);
    }
  }
  
  return same_b;
}

/*
 * static void directory_make
 *
 * DESCRIPTION:
 *
 * Create a directory.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * dir -> Directory to create.
 */
static	void	directory_make(const char *dir)
{
  int	ret;
  
  if (dir[0] == '\0') {
    if (! quiet_b) {
      (void)fprintf(stderr, "%s: bad directory path '/': ", argv_program);
    }
    exit(1);
  }
  
  if (verbose_b || very_verbose_b) {
    (void)printf("Making directory '%s'\n", dir);
  }
  
  if (! execute_cmds_b) {
    return;
  }
  
  ret = mkdir(dir, dir_mode);
  if (ret != 0) {
    /* do we need to make the path? */
    if (errno == ENOENT && dir_makeall_b) {
      make_path(dir);
      ret = mkdir(dir, dir_mode);
    }
    if (ret != 0 && ((! quiet_b) || errno != EEXIST)) {
      if (! quiet_b) {
	(void)fprintf(stderr, "%s: error making directory '%s': %s\n",
		      argv_program, dir, strerror(errno));
      }
      exit(1);
    }
  }
  
  /* set dest. file permissions */
  (void)set_permissions(dir, NULL);
}

int	main(int argc, char **argv)
{
  char		to[MAXPATHLEN];
  char		*from, *dest_path, *path_p, *end_p;
  int		installed_b, did_any_b = 0, dest_is_dir_b = 0, to_exist_b;
  int		euid, file_c, ret;
  struct stat	from_stat, to_stat, *stat_p;
  
  argv_version_string = version_id;
  argv_process(args, argc, argv);
  
  if (help_b || files.aa_entry_n == 0) {
    (void)printf("Install Utility: http://256.com/sources/install/\n");
    (void)printf("  This utility is a replacement for the system's install\n");
    (void)printf("  and provides some useful additional functionality.\n");
    (void)printf("  For a list of the command-line options enter: %s --usage\n",
		 argv_argv[0]);
    exit(0);
  }
  
  /* very_verbose_b wins */
  if (very_verbose_b && verbose_b) {
    verbose_b = 0;
  }
  /* verbose_b wins */
  if (verbose_b && quiet_b) {
    quiet_b = 0;
  }
  /*
   * If we are installing symlinks then forget about comparing the
   * files.  Also, we should not do things like strip or ranlib which
   * will operate on the original.
   */
  if (abs_symlink_files_b || rel_symlink_files_b) {
    examine_b = 0;
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(RANLIB_PROG)
    ranlib_libs_b = 0;
#endif
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(STRIP_PROG)
    strip_bins_b = 0;
#endif
  }
  
  if (dir_make_b) {
    /* should be 1 directory name */
    if (files.aa_entry_n != 1) {
      argv_usage(args, ARGV_USAGE_SHORT);
      exit(1);
    }
  }
  else {
    /* must be at least 1 file and dest */
    if (files.aa_entry_n < 2) {
      argv_usage(args, ARGV_USAGE_SHORT);
      exit(1);
    }
  }
  
  /*
   * HACK: if we specified mkdir and file-mode but not directory-mode
   * then set dir-mode to be file-mode
   */
  if (dir_make_b
      && argv_was_used(args, ARG_MODE)
      && (! argv_was_used(args, ARG_DIR_MODE))) {
    dir_mode = file_mode;
  }
#ifdef HAVE_UMASK
  if (umask_mode < 0) {
    umask(0);
  }
  else {
    (void)umask(umask_mode);
    file_mode &= ~umask_mode;
    dir_mode &= ~umask_mode;
  }
#endif
  
  /*
   * setup variables
   */
  
  /*
   * get owner, group info
   */
  euid = geteuid();
  if (euid == ROOT_UID) {
    char	*dotp;
    
    dotp = strchr(file_owner, '.');
    if (dotp != NULL) {
      if (argv_was_used(args, ARG_GROUP)) {
	(void)fprintf(stderr,
		      "%s: owner '%s' has a dot but you specified a group\n",
		      argv_program, file_owner);
	exit(1);
      }
      *dotp = '\0';
      file_group = dotp + 1;
    }
    
    passwd_p = getpwnam(file_owner);
    if (passwd_p == NULL) {
      (void)fprintf(stderr, "%s: unknown username '%s'\n",
		    argv_program, file_owner);
      exit(1);
    }
  }
  else {
    passwd_p = getpwuid(euid);
    if (passwd_p == NULL) {
      (void)fprintf(stderr, "%s: unknown user id %d\n", argv_program, euid);
      exit(1);
    }
    
    /* if we specified and owner and it is not us, error */
    if (argv_was_used(args, ARG_OWNER)
	&& strcmp(file_owner, passwd_p->pw_name) != 0) {
      (void)fprintf(stderr,
		    "%s: no root permissions, can't specify owner '%s'\n",
		    argv_program, file_owner);
      exit(1);
    }
  }
  
  /* get our group name */
  group_p = getgrnam(file_group);
  if (group_p == NULL) {
    (void)fprintf(stderr, "%s: unknown group '%s'\n",
		  argv_program, file_group);
    exit(1);
  }
  
  /*
   * preprocess destination pathname
   */
  dest_path = ARGV_ARRAY_ENTRY(files, char *, files.aa_entry_n - 1);
  
  /* this will strip off the ending /s */
  end_p = NULL;
  for (path_p = dest_path; *path_p != '\0'; path_p++) {
    if (*path_p != '/') {
      end_p = path_p;
    }
  }
  if (end_p != NULL) {
    *(end_p + 1) = '\0';
  }
  
  /*
   * start the work
   */
  
  /* are we doing a directory? */
  if (dir_make_b) {
    directory_make(dest_path);
    argv_cleanup(args);
    exit(0);
  }
  
  /* test to see if it is a directory or not */
  if ((! dir_makeall_b)
      && (stat(dest_path, &to_stat) != 0 || (! (to_stat.st_mode & S_IFDIR)))) {
    if (files.aa_entry_n > 2) {
      (void)fprintf(stderr,
		    "%s: can't install 2 or more files into a non-directory\n",
		    argv_program);
      exit(1);
    }
    dest_is_dir_b = 0;
  }
  else {
    dest_is_dir_b = 1;
  }
  
  /* install the file(s) */
  for (file_c = 0; file_c < files.aa_entry_n - 1; file_c++) {
    
    from = ARGV_ARRAY_ENTRY(files, char *, file_c);
    
    /*
     * check from file out
     */
    if (stat(from, &from_stat) != 0) {
      (void)fprintf(stderr, "%s: could not stat '%s': %s\n",
		    argv_program, from, strerror(errno));
      exit(1);
    }
    
    if (! ((from_stat.st_mode & S_IFREG)
	   || (from_stat.st_mode & S_IFDIR)
	   || (from_stat.st_mode & S_IFCHR)
	   || (from_stat.st_mode & S_IFBLK)
	   || (from_stat.st_mode & S_IFIFO))) {
      (void)printf("Skipping '%s': not a file, directory, node, or FIFO.\n",
		   from);
      /* if there was one file then exit with an error */
      if (files.aa_entry_n == 1) {
	exit(1);
      }
      continue;
    }
    
    /*
     * build the to-file pathname
     */
    
    /* make 'to' path */
    if (dest_is_dir_b) {
      path_p = strrchr(from, '/');
      if (path_p == NULL) {
	path_p = from;
      }
      else {
	path_p++;
      }
      (void)sprintf(to, "%s/%s", dest_path, path_p);
    }
    else {
      (void)strcpy(to, dest_path);
    }
    
    /* do we need to remove extensions? */
    if (remove_ext_b) {
      /* go forward looking for dot after last / */
      end_p = NULL;
      for (path_p = to; *path_p != '\0'; path_p++) {
	if (*path_p == '.') {
	  end_p = path_p;
	}
	else if (*path_p == '/') {
	  end_p = NULL;
	}
      }
      if (end_p != NULL) {
	*end_p = '\0';
      }
    }
    
    /*
     * check the 'to' file out to see if it the same as 'from'
     */
    ret = stat(to, &to_stat);
    if (ret == 0) {
      stat_p = &to_stat;
      to_exist_b = 1;
    }
    else {
      stat_p = NULL;
      to_exist_b = 0;
#ifdef HAVE_LSTAT
      /* see if we have a symlink instead */
      if (lstat(to, &to_stat) == 0) {
	to_exist_b = 1;
      }
#endif
    }
    
    if (to_exist_b
	&& to_stat.st_dev == from_stat.st_dev
	&& to_stat.st_ino == from_stat.st_ino) {
      if ((! examine_b)
#ifdef HAVE_SYMLINK
	  && (! rel_symlink_files_b)
	  && (! abs_symlink_files_b)
#endif
	  ) {
	(void)printf("Warning: '%s' and '%s' are the same file.\n",
		     from, to);
      }
      else if (very_verbose_b) {
	(void)printf("File '%s' and '%s' are the same file\n", from, to);
      }
      continue;
    }
    
    installed_b = 0;
    
    /* are the files exactly the same? */
    if (! (to_exist_b
	   && examine_b
	   && are_same(from, to)
	   && (to_stat.st_uid == euid || euid == ROOT_UID)
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(RANLIB_PROG)
	   && (! ranlib_libs_b)
#endif
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(STRIP_PROG)
	   && (! strip_bins_b)
#endif
	   )) {
      
      /* install the file., do the linking / copying */
      if (install_file(from, to, to_exist_b) != 1) {
	continue;
      }
      
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(RANLIB_PROG)
      /* run utility programs on destination if necessary and available */
      if (ranlib_libs_b) {
	(void)run_command(RANLIB_PROG, to);
      }
#endif
#if defined(HAVE_FORK) && defined(HAVE_EXECLP) && defined(STRIP_PROG)
      if (strip_bins_b) {
	(void)run_command(STRIP_PROG, to);
      }
#endif
      
      stat_p = NULL;
      installed_b = 1;
      did_any_b = 1;
    }
    
    /* if we installed the files or we need to update the target's modes */
    if (installed_b
	|| (! ((to_stat.st_mode & 07777) == file_mode
	       && to_stat.st_gid == group_p->gr_gid
	       && to_stat.st_uid == passwd_p->pw_uid))) {
      if (set_permissions(to, stat_p)) {
	/* print a message if we did not get a message during install */
	if (verbose_b && (! installed_b)) {
	  (void)printf("Changing '%s' modes to '%s.%s' and %#o\n",
		       to, passwd_p->pw_name, group_p->gr_name, file_mode);
	}
      }
    }
    
    /* do we need to check after? */
    if (compare_b
#ifdef HAVE_SYMLINK
	&& (! rel_symlink_files_b)
	&& (! abs_symlink_files_b)
#endif
	&& (! are_same(from, to))) {
      if (! quiet_b) {
	(void)fprintf(stderr,
		      "%s: file '%s' is not the same after install as '%s'\n",
		      argv_program, from, to);
      }
      exit(1);
    }
  }
  
  argv_cleanup(args);
  
  if (error_code_b && (! did_any_b)) {
    if (verbose_b) {
      (void)printf("%s: Error.  Installed no files.\n", argv_program);
    }
    exit(1);
  }
  
  exit(0);
}
